@@ -1,30 +1,33 @@
 """Accessing data from the supported databases through their APIs."""
+from __future__ import annotations
+
 import io
 import itertools
+import warnings
 import zipfile
-from collections import OrderedDict
-from numbers import Number
 from pathlib import Path
-from ssl import SSLContext
-from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, Union
+from typing import TYPE_CHECKING, Any, Mapping, Sequence, Tuple, Union, cast
 from unittest.mock import patch
 
 import async_retriever as ar
-import cytoolz as tlz
+import cytoolz.curried as tlz
+import dask.config
 import geopandas as gpd
 import numpy as np
 import pandas as pd
+import pygeoogc as ogc
 import pygeoutils as geoutils
 import pyproj
 import rasterio as rio
 import xarray as xr
 from pygeoogc import WMS, ArcGISRESTful, RetrySession, ServiceURL
 from pygeoogc import utils as ogc_utils
-from pynhd import AGRBase
-from shapely.geometry import MultiPolygon, Polygon
+from pynhd.core import ScienceBase
+from rioxarray import _io as rxr
 
-from . import helpers
-from .exceptions import (
+from pygeohydro import helpers
+from pygeohydro.exceptions import (
+    DependencyError,
     InputTypeError,
     InputValueError,
     MissingColumnError,
@@ -32,12 +35,16 @@ from .exceptions import (
     ServiceUnavailableError,
     ZeroMatchedError,
 )
-from .helpers import Stats, logger
+from pygeohydro.helpers import Stats
 
-DEF_CRS = "epsg:4326"
-EXPIRE = -1
-GTYPE = Union[Polygon, MultiPolygon, Tuple[float, float, float, float]]
+if TYPE_CHECKING:
+    from numbers import Number
+    from ssl import SSLContext
 
+    from shapely import MultiPolygon, Polygon
+
+    GTYPE = Union[Polygon, MultiPolygon, Tuple[float, float, float, float]]
+    CRSTYPE = Union[int, str, pyproj.CRS]
 
 __all__ = [
     "get_camels",
@@ -47,12 +54,14 @@ __all__ = [
     "nlcd_bycoords",
     "cover_statistics",
     "overland_roughness",
+    "nlcd_area_percent",
+    "soil_properties",
+    "soil_gnatsgo",
     "NID",
-    "WBD",
 ]
 
 
-def get_camels() -> Tuple[gpd.GeoDataFrame, xr.Dataset]:
+def get_camels() -> tuple[gpd.GeoDataFrame, xr.Dataset]:
     """Get streaflow and basin attributes of all 671 stations in CAMELS dataset.
 
     Notes
@@ -66,10 +75,10 @@ def get_camels() -> Tuple[gpd.GeoDataFrame, xr.Dataset]:
         is streamflow data and basin attributes as an ``xarray.Dataset``.
     """
     base_url = "/".join(
-        [
+        (
             "https://thredds.hydroshare.org/thredds/fileServer/hydroshare",
             "resources/658c359b8c83494aac0f58145b1b04e6/data/contents",
-        ]
+        )
     )
     urls = [
         f"{base_url}/camels_attributes_v2.0.feather",
@@ -78,14 +87,32 @@ def get_camels() -> Tuple[gpd.GeoDataFrame, xr.Dataset]:
     resp = ar.retrieve_binary(urls)
 
     attrs = gpd.read_feather(io.BytesIO(resp[0]))
-    qobs = xr.open_dataset(io.BytesIO(resp[1]), engine="h5netcdf")  # type: ignore
+    qobs = xr.open_dataset(io.BytesIO(resp[1]), engine="h5netcdf")
+    qobs["discharge"] = xr.where(qobs["discharge"] < 0, np.nan, qobs["discharge"])
+
+    url = "/".join(
+        (
+            "https://gist.githubusercontent.com/cheginit",
+            "229c83c89eee3801a586bcb3ebb4e825/raw/newman_ids.txt",
+        )
+    )
+    ids = ar.retrieve_text([url])[0].split(",")
+    attrs["Newman_2017"] = attrs.index.isin(ids)
+    qobs["Newman_2017"] = qobs.station_id.isin(ids)
+    desc = ". ".join(
+        (
+            "Whether station is in Newman et al. (2017) dataset",
+            "https://doi.org/10.1175/JHM-D-16-0284.1",
+        )
+    )
+    qobs["Newman_2017"].attrs["description"] = desc
     return attrs, qobs
 
 
 def ssebopeta_bycoords(
     coords: pd.DataFrame,
-    dates: Union[Tuple[str, str], int, List[int]],
-    crs: str = DEF_CRS,
+    dates: tuple[str, str] | int | list[int],
+    crs: CRSTYPE = 4326,
 ) -> xr.Dataset:
     """Daily actual ET for a dataframe of coords from SSEBop database in mm/day.
 
@@ -95,8 +122,8 @@ def ssebopeta_bycoords(
         A dataframe with ``id``, ``x``, ``y`` columns.
     dates : tuple or list, optional
         Start and end dates as a tuple (start, end) or a list of years [2001, 2010, ...].
-    crs : str, optional
-        The CRS of the input coordinates, defaults to epsg:4326.
+    crs : str, int, or pyproj.CRS, optional
+        The CRS of the input coordinates, defaults to ``epsg:4326``.
 
     Returns
     -------
@@ -114,7 +141,7 @@ def ssebopeta_bycoords(
     _coords = gpd.GeoSeries(
         gpd.points_from_xy(coords["x"], coords["y"]), index=coords["id"], crs=crs
     )
-    _coords = _coords.to_crs(DEF_CRS)
+    _coords = _coords.to_crs(4326)
     co_list = list(zip(_coords.x, _coords.y))
 
     f_list = helpers.get_ssebopeta_urls(dates)
@@ -122,7 +149,7 @@ def ssebopeta_bycoords(
 
     with patch("socket.has_ipv6", False):
 
-        def _ssebop(url: str) -> List[np.ndarray]:  # type: ignore
+        def _ssebop(url: str) -> list[np.ndarray]:  # type: ignore
             r = session.get(url)
             z = zipfile.ZipFile(io.BytesIO(r.content))
 
@@ -157,8 +184,8 @@ def ssebopeta_bycoords(
 
 def ssebopeta_bygeom(
     geometry: GTYPE,
-    dates: Union[Tuple[str, str], int, List[int]],
-    geo_crs: str = DEF_CRS,
+    dates: tuple[str, str] | int | list[int],
+    geo_crs: CRSTYPE = 4326,
 ) -> xr.DataArray:
     """Get daily actual ET for a region from SSEBop database.
 
@@ -176,8 +203,8 @@ def ssebopeta_bygeom(
         the order should be (west, south, east, north).
     dates : tuple or list, optional
         Start and end dates as a tuple (start, end) or a list of years [2001, 2010, ...].
-    geo_crs : str, optional
-        The CRS of the input geometry, defaults to epsg:4326.
+    geo_crs : str, int, or pyproj.CRS, optional
+        The CRS of the input geometry, defaults to ``epsg:4326``.
 
     Returns
     -------
@@ -185,29 +212,32 @@ def ssebopeta_bygeom(
         Daily actual ET within a geometry in mm/day at 1 km resolution
     """
     f_list = helpers.get_ssebopeta_urls(dates)
-    if isinstance(geometry, (Polygon, MultiPolygon)):
-        gtiff2xarray = tlz.partial(geoutils.gtiff2xarray, geometry=geometry, geo_crs=geo_crs)
-    else:
-        gtiff2xarray = tlz.partial(geoutils.gtiff2xarray)
 
+    try:
+        _ = geoutils.geo2polygon(geometry)
+    except geoutils.InputTypeError as ex:
+        raise InputTypeError("geometry", "(Multi)Polygon or tuple of length 4") from ex
+
+    gtiff2xarray = tlz.partial(geoutils.gtiff2xarray, geometry=geometry, geo_crs=geo_crs)
     session = RetrySession()
-
     with patch("socket.has_ipv6", False):
 
         def _ssebop(t: pd.Timestamp, url: str) -> xr.DataArray:
             resp = session.get(url)
             zfile = zipfile.ZipFile(io.BytesIO(resp.content))
             content = zfile.read(zfile.filelist[0].filename)
-            ds: xr.DataArray = gtiff2xarray(r_dict={"eta": content})
+            ds = gtiff2xarray(r_dict={"eta": content})
+            ds = cast("xr.DataArray", ds)
             return ds.expand_dims({"time": [t]})
 
-        data = xr.merge(_ssebop(t, url) for t, url in f_list)
-    eta: xr.DataArray = data.where(data.eta < data.eta.nodatavals[0]).eta.copy() * 1e-3
+        data = xr.merge(itertools.starmap(_ssebop, f_list))
+    eta: xr.DataArray = data.eta
+    eta = eta.where(eta < eta.rio.nodata) * 1e-3
     eta.attrs.update(
         {
             "units": "mm/day",
             "nodatavals": (np.nan,),
-            "crs": DEF_CRS,
+            "crs": 4326,
             "long_name": "Actual ET",
         }
     )
@@ -225,9 +255,10 @@ class NLCD:
         Layers that are not in years are ignored, e.g., ``{'cover': [2016, 2019]}`` returns
         land cover data for 2016 and 2019.
     region : str, optional
-        Region in the US, defaults to ``L48``. Valid values are L48 (for CONUS), HI (for Hawaii),
-        AK (for Alaska), and PR (for Puerto Rico). Both lower and upper cases are acceptable.
-    crs : str, optional
+        Region in the US that the input geometries are located, defaults to ``L48``.
+        Valid values are ``L48`` (for CONUS), ``HI`` (for Hawaii), ``AK`` (for Alaska),
+        and ``PR`` (for Puerto Rico). Both lower and upper cases are acceptable.
+    crs : str, int, or pyproj.CRS, optional
         The spatial reference system to be used for requesting the data, defaults to
         ``epsg:4326``.
     ssl : bool or SSLContext, optional
@@ -237,10 +268,10 @@ class NLCD:
 
     def __init__(
         self,
-        years: Optional[Mapping[str, Union[int, List[int]]]] = None,
+        years: Mapping[str, int | list[int]] | None = None,
         region: str = "L48",
-        crs: str = DEF_CRS,
-        ssl: Union[SSLContext, bool, None] = None,
+        crs: CRSTYPE = 4326,
+        ssl: SSLContext | bool | None = None,
     ) -> None:
         default_years = {
             "impervious": [2019],
@@ -258,25 +289,9 @@ class NLCD:
         if self.crs not in self.valid_crs:
             raise InputValueError("crs", self.valid_crs)
         self.layers = self.get_layers()
-        self.units = OrderedDict(
-            (
-                ("impervious", "%"),
-                ("cover", "classes"),
-                ("canopy", "%"),
-                ("descriptor", "classes"),
-            )
-        )
-        self.types = OrderedDict(
-            (
-                ("impervious", "f4"),
-                ("cover", "u1"),
-                ("canopy", "f4"),
-                ("descriptor", "u1"),
-            )
-        )
-        self.nodata = OrderedDict(
-            (("impervious", np.nan), ("cover", 127), ("canopy", np.nan), ("descriptor", 127))
-        )
+        self.units = {"impervious": "%", "cover": "classes", "canopy": "%", "descriptor": "classes"}
+        self.types = {"impervious": "f4", "cover": "u1", "canopy": "f4", "descriptor": "u1"}
+        self.nodata = {"impervious": np.nan, "cover": 127, "canopy": np.nan, "descriptor": 127}
 
         self.wms = WMS(
             ServiceURL().wms.mrlc,
@@ -287,7 +302,7 @@ class NLCD:
             ssl=ssl,
         )
 
-    def get_layers(self) -> List[str]:
+    def get_layers(self) -> list[str]:
         """Get NLCD layers for the provided years dictionary."""
         valid_regions = ["L48", "HI", "PR", "AK"]
         if self.region not in valid_regions:
@@ -321,31 +336,22 @@ class NLCD:
             for yr in yrs
         ]
 
-    def get_response(
-        self, bbox: Tuple[float, float, float, float], resolution: float
-    ) -> Dict[str, bytes]:
-        """Get response from a url."""
-        return self.wms.getmap_bybox(bbox, resolution, self.crs)
-
-    def to_xarray(
+    def get_map(
         self,
-        r_dict: Dict[str, bytes],
-        geometry: Union[Polygon, MultiPolygon, None] = None,
+        geometry: Polygon | MultiPolygon,
+        resolution: int,
     ) -> xr.Dataset:
-        """Convert response to xarray.DataArray."""
-        if isinstance(geometry, (Polygon, MultiPolygon)):
-            gtiff2xarray = tlz.partial(
-                geoutils.gtiff2xarray, geometry=geometry, geo_crs=self.crs, nodata=255
-            )
-        else:
-            gtiff2xarray = tlz.partial(geoutils.gtiff2xarray, nodata=255)
-
+        """Get NLCD response and convert it to ``xarray.DataArray``."""
+        r_dict = self.wms.getmap_bybox(geometry.bounds, resolution, self.crs)
+        gtiff2xarray = tlz.partial(
+            geoutils.gtiff2xarray, geometry=geometry, geo_crs=self.crs, nodata=255
+        )
         try:
             _ds = gtiff2xarray(r_dict=r_dict)
         except rio.RasterioIOError as ex:
             raise ServiceUnavailableError(self.wms.url) from ex
 
-        ds: xr.Dataset = _ds.to_dataset() if isinstance(_ds, xr.DataArray) else _ds
+        ds = _ds.to_dataset() if isinstance(_ds, xr.DataArray) else _ds
         ds.attrs = _ds.attrs
         for lyr in self.layers:
             name = [n for n in self.units if n in lyr.lower()][-1]
@@ -364,13 +370,13 @@ class NLCD:
 
 
 def nlcd_bygeom(
-    geometry: Union[gpd.GeoSeries, gpd.GeoDataFrame],
-    resolution: float,
-    years: Optional[Mapping[str, Union[int, List[int]]]] = None,
+    geometry: gpd.GeoSeries | gpd.GeoDataFrame,
+    resolution: int,
+    years: Mapping[str, int | list[int]] | None = None,
     region: str = "L48",
-    crs: str = DEF_CRS,
-    ssl: Union[SSLContext, bool, None] = None,
-) -> Dict[int, xr.Dataset]:
+    crs: CRSTYPE = 4326,
+    ssl: SSLContext | bool | None = None,
+) -> dict[int | str, xr.Dataset]:
     """Get data from NLCD database (2019).
 
     Parameters
@@ -387,12 +393,12 @@ def nlcd_bygeom(
         Layers that are not in years are ignored, e.g., ``{'cover': [2016, 2019]}`` returns
         land cover data for 2016 and 2019.
     region : str, optional
-        Region in the US, defaults to ``L48``. Valid values are ``L48`` (for CONUS),
-        ``HI`` (for Hawaii), ``AK`` (for Alaska), and ``PR`` (for Puerto Rico).
-        Both lower and upper cases are acceptable.
-    crs : str, optional
+        Region in the US that the input geometries are located, defaults to ``L48``.
+        Valid values are ``L48`` (for CONUS), ``HI`` (for Hawaii), ``AK`` (for Alaska),
+        and ``PR`` (for Puerto Rico). Both lower and upper cases are acceptable.
+    crs : str, int, or pyproj.CRS, optional
         The spatial reference system to be used for requesting the data, defaults to
-        epsg:4326.
+        ``epsg:4326``.
     ssl : bool or SSLContext, optional
         SSLContext to use for the connection, defaults to None. Set to ``False`` to disable
         SSL certification verification.
@@ -404,7 +410,11 @@ def nlcd_bygeom(
         of the input ``GeoDataFrame``.
     """
     if resolution < 30:
-        logger.warning("NLCD's resolution is 30 m, so finer resolutions are not recommended.")
+        warnings.warn(
+            "NLCD's resolution is 30 m, so finer resolutions are not recommended.",
+            UserWarning,
+            stacklevel=2,
+        )
 
     if not isinstance(geometry, (gpd.GeoDataFrame, gpd.GeoSeries)):
         raise InputTypeError("geometry", "GeoDataFrame or GeoSeries")
@@ -415,18 +425,14 @@ def nlcd_bygeom(
 
     nlcd_wms = NLCD(years=years, region=region, crs=crs, ssl=ssl)
 
-    ds = {
-        i: nlcd_wms.to_xarray(nlcd_wms.get_response(g.bounds, resolution), g)
-        for i, g in _geometry.items()
-    }
-    return ds
+    return {i: nlcd_wms.get_map(g, resolution) for i, g in _geometry.items()}
 
 
 def nlcd_bycoords(
-    coords: List[Tuple[float, float]],
-    years: Optional[Mapping[str, Union[int, List[int]]]] = None,
+    coords: list[tuple[float, float]],
+    years: Mapping[str, int | list[int]] | None = None,
     region: str = "L48",
-    ssl: Union[SSLContext, bool, None] = None,
+    ssl: SSLContext | bool | None = None,
 ) -> gpd.GeoDataFrame:
     """Get data from NLCD database (2019).
 
@@ -440,9 +446,9 @@ def nlcd_bycoords(
         Layers that are not in years are ignored, e.g., ``{'cover': [2016, 2019]}`` returns
         land cover data for 2016 and 2019.
     region : str, optional
-        Region in the US, defaults to ``L48``. Valid values are ``L48`` (for CONUS),
-        ``HI`` (for Hawaii), ``AK`` (for Alaska), and ``PR`` (for Puerto Rico).
-        Both lower and upper cases are acceptable.
+        Region in the US that the input geometries are located, defaults to ``L48``.
+        Valid values are ``L48`` (for CONUS), ``HI`` (for Hawaii), ``AK`` (for Alaska),
+        and ``PR`` (for Puerto Rico). Both lower and upper cases are acceptable.
     ssl : bool or SSLContext, optional
         SSLContext to use for the connection, defaults to None. Set to ``False`` to disable
         SSL certification verification.
@@ -456,10 +462,10 @@ def nlcd_bycoords(
         raise InputTypeError("coords", "list of (lon, lat)")
 
     nlcd_wms = NLCD(years=years, region=region, crs="epsg:3857", ssl=ssl)
-    points = gpd.GeoSeries(gpd.points_from_xy(*zip(*coords)), crs=DEF_CRS)
+    points = gpd.GeoSeries(gpd.points_from_xy(*zip(*coords)), crs=4326)
     points_proj = points.to_crs(nlcd_wms.crs)
-    bounds = points_proj.buffer(50, cap_style=3)
-    ds_list = [nlcd_wms.to_xarray(nlcd_wms.get_response(b.bounds, 30)) for b in bounds]
+    geoms = points_proj.buffer(50, cap_style=3)
+    ds_list = [nlcd_wms.get_map(g, 30) for g in geoms]
 
     def get_value(da: xr.DataArray, x: float, y: float) -> Number:
         nodata = da.attrs["nodatavals"][0]
@@ -541,47 +547,157 @@ def cover_statistics(cover_da: xr.DataArray) -> Stats:
     return Stats(class_percentage, category_percentage)
 
 
+def _area_percent(nlcd: xr.Dataset, year: int) -> dict[str, float]:
+    """Calculate the percentage of the area for each land cover class."""
+    cover_nodata = nlcd[f"cover_{year}"].rio.nodata
+    if np.isnan(cover_nodata):
+        msk = ~nlcd[f"cover_{year}"].isnull()
+    elif cover_nodata > 0:
+        msk = nlcd[f"cover_{year}"] < cover_nodata
+    else:
+        msk = nlcd[f"cover_{year}"] > cover_nodata
+    cell_total = msk.sum()
+
+    msk = nlcd[f"cover_{year}"].isin(range(21, 25))
+    urban = msk.sum() / cell_total
+    natural = 1 - urban
+
+    impervious = nlcd.where(msk)[f"impervious_{year}"].mean() * urban / 100
+    developed = urban - impervious
+    natural = natural.compute().item() * 100
+    developed = developed.compute().item() * 100
+    impervious = impervious.compute().item() * 100
+    return {
+        "natural": natural,
+        "developed": developed,
+        "impervious": impervious,
+        "urban": developed + impervious,
+    }
+
+
+def nlcd_area_percent(
+    geo_df: gpd.GeoSeries | gpd.GeoDataFrame,
+    year: int = 2019,
+    region: str = "L48",
+) -> pd.DataFrame:
+    """Compute the area percentages of the natural, developed, and impervious areas.
+
+    Notes
+    -----
+    This function uses imperviousness and land use/land cover data from NLCD
+    to compute the area percentages of the natural, developed, and impervious areas.
+    It considers land cover classes of 21 to 24 as urban and the rest as natural.
+    Then, uses imperviousness percentage to partition the urban area into developed
+    and impervious areas. So, ``urban = developed + impervious`` and always
+    ``natural + urban = natural + developed + impervious = 100``.
+
+    Parameters
+    ----------
+    geometry : geopandas.GeoDataFrame or geopandas.GeoSeries
+        A GeoDataFrame or GeoSeries with the geometry to query. The indices are used
+        as keys in the output dictionary.
+    year : int, optional
+        Year of the NLCD data, defaults to 2019. Available years are 2019, 2016, 2013,
+        2011, 2008, 2006, 2004, and 2001.
+    region : str, optional
+        Region in the US that the input geometries are located, defaults to ``L48``.
+        Valid values are ``L48`` (for CONUS), ``HI`` (for Hawaii), ``AK`` (for Alaska),
+        and ``PR`` (for Puerto Rico). Both lower and upper cases are acceptable.
+
+    Returns
+    -------
+    pandas.DataFrame
+        A dataframe with the same index as input ``geo_df`` and columns are the area
+        percentages of the natural, developed, impervious, and urban
+        (sum of developed and impervious) areas. Sum of urban and natural percentages
+        is always 100, as well as the sume of natural, developed, and impervious
+        percentages.
+    """
+    valid_year = (2019, 2016, 2013, 2011, 2008, 2006, 2004, 2001)
+    if year not in valid_year:
+        raise InputValueError("year", valid_year)
+
+    if not isinstance(geo_df, (gpd.GeoDataFrame, gpd.GeoSeries)):
+        raise InputTypeError("geometry", "GeoDataFrame or GeoSeries")
+
+    if geo_df.crs is None:
+        raise MissingCRSError
+
+    geoms = geo_df.to_crs(4326).geometry
+
+    wms = NLCD(years={"impervious": year, "cover": year}, region=region)
+
+    return pd.DataFrame.from_dict(
+        {i: _area_percent(wms.get_map(g, 30), year) for i, g in geoms.items()},
+        orient="index",
+    )
+
+
 class NID:
     """Retrieve data from the National Inventory of Dams web service."""
 
     def __init__(self) -> None:
         self.base_url = ServiceURL().restful.nid
         self.suggest_url = f"{self.base_url}/suggestions"
-        self.fields_meta = pd.DataFrame(ar.retrieve_json([f"{self.base_url}/advanced-fields"])[0])
+        resp = ar.retrieve_json([f"{self.base_url}/advanced-fields"])
+        resp = cast("list[dict[str, Any]]", resp)
+        self.fields_meta = pd.DataFrame(resp[0])
         self.valid_fields = self.fields_meta.name
         self.dam_type = {
-            -1: "N/A",
+            pd.NA: "N/A",
+            None: "N/A",
             1: "Arch",
-            2: "Buttress",
-            3: "Concrete",
-            4: "Earth",
-            5: "Gravity",
-            6: "Masonry",
-            7: "Multi-Arch",
-            8: "Rockfill",
-            9: "Roller-Compacted Concrete",
-            10: "Stone",
-            11: "Timber Crib",
+            2: "Multi-Arch",
+            3: "Stone",
+            4: "Roller-Compacted Concrete",
+            5: "Rockfill",
+            6: "Buttress",
+            7: "Masonry",
+            8: "Earth",
+            9: "Gravity",
+            10: "Timber Crib",
+            11: "Concrete",
             12: "Other",
         }
+        self.dam_type.update({str(k): v for k, v in self.dam_type.items() if str(k).isdigit()})
         self.dam_purpose = {
-            -1: "N/A",
-            1: "Debris Control",
-            2: "Fire Protection, Stock, Or Small Farm Pond",
-            3: "Fish and Wildlife Pond",
-            4: "Flood Risk Reduction",
-            5: "Grade Stabilization",
+            pd.NA: "N/A",
+            None: "N/A",
+            1: "Tailings",
+            2: "Irrigation",
+            3: "Navigation",
+            4: "Fish and Wildlife Pond",
+            5: "Recreation",
             6: "Hydroelectric",
-            7: "Irrigation",
-            8: "Navigation",
-            9: "Recreation",
-            10: "Tailings",
-            11: "Water Supply",
+            7: "Debris Control",
+            8: "Water Supply",
+            9: "Flood Risk Reduction",
+            10: "Fire Protection, Stock, Or Small Farm Pond",
+            11: "Grade Stabilization",
             12: "Other",
         }
+        self.dam_purpose.update(
+            {str(k): v for k, v in self.dam_purpose.items() if str(k).isdigit()}
+        )
+        self.data_units = {
+            "distance": "mile",
+            "damHeight": "ft",
+            "hydraulicHeight": "ft",
+            "structuralHeight": "ft",
+            "nidHeight": "ft",
+            "damLength": "ft",
+            "volume": "cubic yards",
+            "nidStorage": "acre-ft",
+            "maxStorage": "acre-ft",
+            "normalStorage": "acre-ft",
+            "surfaceArea": "acre",
+            "drainageArea": "square miles",
+            "maxDischarge": "cfs",
+            "spillwayWidth": "ft",
+        }
         self.nid_inventory_path = Path("cache", "nid_inventory.feather")
 
-    def stage_nid_inventory(self, fname: Union[str, Path, None] = None) -> None:
+    def stage_nid_inventory(self, fname: str | Path | None = None) -> None:
         """Download the entire NID inventory data and save to a feather file.
 
         Parameters
@@ -602,7 +718,7 @@ class NID:
                     "national-export/dams/nation/7783878c-6db4-46aa-99ef-8c283109ea78/nation.gpkg",
                 ]
             )
-            ar.stream_write([url], [fname.with_suffix(".gpkg")])
+            _ = ogc.streaming_download(url, fnames=fname.with_suffix(".gpkg"))
             dams = gpd.read_file(fname.with_suffix(".gpkg"))
             dams = dams.astype(
                 {
@@ -702,8 +818,8 @@ class NID:
 
     @staticmethod
     def _get_json(
-        urls: Sequence[str], params: Optional[List[Dict[str, str]]] = None
-    ) -> List[Dict[str, Any]]:
+        urls: Sequence[str], params: list[dict[str, str]] | None = None
+    ) -> list[dict[str, Any]]:
         """Get JSON response from NID web service.
 
         Parameters
@@ -726,18 +842,22 @@ class NID:
         else:
             kwds = [{"params": {**p, "out": "json"}} for p in params]
         resp = ar.retrieve_json(urls, kwds)
-        if len(resp) == 0:
+        resp = cast("list[dict[str, Any]]", resp)
+        if not resp:
             raise ZeroMatchedError
 
         failed = [(i, f"Req_{i}: {r['message']}") for i, r in enumerate(resp) if "error" in r]
         if failed:
             idx, err_msgs = zip(*failed)
+            idx = cast("Tuple[int]", idx)
+            err_msgs = cast("Tuple[str]", err_msgs)
             errs = " service requests failed with the following messages:\n"
             errs += "\n".join(err_msgs)
             if len(failed) == len(urls):
                 raise ZeroMatchedError(f"All{errs}")
             resp = [r for i, r in enumerate(resp) if i not in idx]
-            logger.warning(f"Some{errs}")
+            fail_count = f"{len(failed)} of {len(urls)}"
+            warnings.warn(f"{fail_count}{errs}", UserWarning, stacklevel=2)
         return resp
 
     @staticmethod
@@ -757,10 +877,10 @@ class NID:
         return gpd.GeoDataFrame(
             nid_df,
             geometry=gpd.points_from_xy(nid_df.longitude, nid_df.latitude),
-            crs=DEF_CRS,
+            crs=4326,
         )
 
-    def get_byfilter(self, query_list: List[Dict[str, List[str]]]) -> List[gpd.GeoDataFrame]:
+    def get_byfilter(self, query_list: list[dict[str, list[str]]]) -> list[gpd.GeoDataFrame]:
         """Query dams by filters from the National Inventory of Dams web service.
 
         Parameters
@@ -802,7 +922,7 @@ class NID:
             for r in self._get_json([f"{self.base_url}/query"] * len(params), params)
         ]
 
-    def get_bygeom(self, geometry: GTYPE, geo_crs: str) -> gpd.GeoDataFrame:
+    def get_bygeom(self, geometry: GTYPE, geo_crs: CRSTYPE) -> gpd.GeoDataFrame:
         """Retrieve NID data within a geometry.
 
         Parameters
@@ -810,7 +930,7 @@ class NID:
         geometry : Polygon, MultiPolygon, or tuple of length 4
             Geometry or bounding box (west, south, east, north) for extracting the data.
         geo_crs : list of str
-            The CRS of the input geometry, defaults to epsg:4326.
+            The CRS of the input geometry, defaults to ``epsg:4326``.
 
         Returns
         -------
@@ -821,11 +941,11 @@ class NID:
         --------
         >>> from pygeohydro import NID
         >>> nid = NID()
-        >>> dams = nid.get_bygeom((-69.77, 45.07, -69.31, 45.45), "epsg:4326")
+        >>> dams = nid.get_bygeom((-69.77, 45.07, -69.31, 45.45), 4326)
         >>> print(dams.name.iloc[0])
         Little Moose
         """
-        _geometry = geoutils.geo2polygon(geometry, geo_crs, DEF_CRS)
+        _geometry = geoutils.geo2polygon(geometry, geo_crs, 4326)
         wbd = ArcGISRESTful(ServiceURL().restful.wbd, 4, outformat="json", outfields="huc8")
         resp = wbd.get_features(wbd.oids_bygeom(_geometry), return_geom=False)
         huc_ids = [
@@ -835,7 +955,7 @@ class NID:
         dams = self.get_byfilter([{"huc8": huc_ids}])[0]
         return dams[dams.within(_geometry)].copy()
 
-    def inventory_byid(self, dam_ids: List[int], stage_nid: bool = False) -> gpd.GeoDataFrame:
+    def inventory_byid(self, federal_ids: list[int], stage_nid: bool = False) -> gpd.GeoDataFrame:
         """Get extra attributes for dams based on their dam ID.
 
         Notes
@@ -848,9 +968,8 @@ class NID:
 
         Parameters
         ----------
-        dam_ids : list of int or str
-            List of the target dam IDs (digists only). Note that the dam IDs are not the
-            same as the NID IDs.
+        federal_ids : list of str
+            List of the target dam Federal IDs.
         stage_nid : bool, optional
             Whether to get the entire NID and then query locally or query from the
             NID web service which tends to be very slow for large number of requests.
@@ -867,27 +986,26 @@ class NID:
         --------
         >>> from pygeohydro import NID
         >>> nid = NID()
-        >>> dams = nid.inventory_byid([514871, 459170, 514868, 463501, 463498])
+        >>> dams = nid.inventory_byid(['KY01232', 'GA02400', 'NE04081', 'IL55070', 'TN05345'])
         >>> print(dams.damHeight.max())
-        120.0
+        39.0
         """
-        urls = [
-            f"{self.base_url}/dams/{i}/inventory"
-            for i in itertools.takewhile(lambda x: str(x).isdigit(), dam_ids)
-        ]
-        if len(urls) != len(dam_ids):
-            raise InputTypeError("dam_ids", "list of digits")
+        ids = set(federal_ids) if isinstance(federal_ids, (list, tuple)) else {federal_ids}
+        ids = {str(i).upper() for i in ids}
+        urls = [f"{self.base_url}/dams/{i}/inventory" for i in ids]
+        if len(urls) != len(ids):
+            raise InputTypeError("dam_ids", "list of Federal IDs")
 
         if stage_nid:
             self.stage_nid_inventory()
             dams = gpd.read_feather(self.nid_inventory_path)
-            return dams[dams.id.isin(tlz.map(int, dam_ids))].copy()
+            return dams[dams.federalId.isin(list(ids))].copy()
 
-        return self._to_geodf(pd.DataFrame(self._get_json(urls)))
+        return self._to_geodf(pd.DataFrame(self._get_json(urls)).set_index("id"))
 
     def get_suggestions(
-        self, text: str, context_key: str = ""
-    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
+        self, text: str, context_key: str | None = None
+    ) -> tuple[pd.DataFrame, pd.DataFrame]:
         """Get suggestions from the National Inventory of Dams web service.
 
         Notes
@@ -914,16 +1032,17 @@ class NID:
         --------
         >>> from pygeohydro import NID
         >>> nid = NID()
-        >>> dams, contexts = nid.get_suggestions("texas", "city")
-        >>> print(contexts.loc["CITY", "value"])
-        Texas City
+        >>> dams, contexts = nid.get_suggestions("houston", "city")
+        >>> print(contexts["suggestion"].to_list())
+        ['Houston', 'Houston Lake']
         """
         fields = self.valid_fields.to_list()
-        if len(context_key) > 0 and context_key not in fields:
-            raise InputValueError("context", fields)
-
-        params = [{"text": text, "contextKey": context_key}]
-        resp = self._get_json([f"{self.base_url}/suggestions"] * len(params), params)
+        params = {"text": text}
+        if context_key:
+            if context_key not in fields:
+                raise InputValueError("context", fields)
+            params["contextKey"] = context_key
+        resp = self._get_json([f"{self.base_url}/suggestions"], [params])
         dams = pd.DataFrame(resp[0]["dams"])
         contexts = pd.DataFrame(resp[0]["contexts"])
         return (
@@ -944,49 +1063,125 @@ class NID:
         )
 
 
-class WBD(AGRBase):
-    """Access Watershed Boundary Dataset (WBD).
+def soil_properties(
+    properties: list[str] | str = "*", soil_dir: str | Path = "cache"
+) -> xr.Dataset:
+    """Get soil properties dataset in the United States from ScienceBase.
 
     Notes
     -----
-    This web service offers Hydrologic Unit (HU) polygon boundaries for
-    the United States, Puerto Rico, and the U.S. Virgin Islands.
-    For more info visit: https://hydro.nationalmap.gov/arcgis/rest/services/wbd/MapServer
+    This function downloads the source zip files from
+    `ScienceBase <https://www.sciencebase.gov/catalog/item/5fd7c19cd34e30b9123cb51f>`__
+    , extracts the included ``.tif`` files, and return them as an ``xarray.Dataset``.
 
     Parameters
     ----------
-    layer : str, optional
-        A valid service layer. Valid layers are:
-
-        - ``wbdline``
-        - ``huc2``
-        - ``huc4``
-        - ``huc6``
-        - ``huc8``
-        - ``huc10``
-        - ``huc12``
-        - ``huc14``
-        - ``huc16``
-
-    outfields : str or list, optional
-        Target field name(s), default to "*" i.e., all the fields.
-    crs : str, optional
-        Target spatial reference, default to ``EPSG:4326``.
+    properties : list of str or str, optional
+        Soil properties to extract, default to "*", i.e., all the properties.
+        Available properties are ``awc`` for available water capacity, ``fc`` for
+        field capacity, and ``por`` for porosity.
+    soil_dir : str or pathlib.Path
+        Directory to store zip files or if exists read from them, defaults to
+        ``./cache``.
     """
+    valid_props = {
+        "awc": {"name": "awc", "units": "mm/m", "long_name": "Available Water Capacity"},
+        "fc": {"name": "fc", "units": "mm/m", "long_name": "Field Capacity"},
+        "por": {"name": "porosity", "units": "mm/m", "long_name": "Porosity"},
+    }
+    if properties == "*":
+        prop = list(valid_props)
+    else:
+        prop = [properties] if isinstance(properties, str) else properties
+        if not all(p in valid_props for p in prop):
+            raise InputValueError("properties", list(valid_props))
+    sb = ScienceBase()
+    soil = sb.get_file_urls("5fd7c19cd34e30b9123cb51f")
+    pat = "|".join(prop)
+    _files, urls = zip(*soil[soil.index.str.contains(f"{pat}.*zip")].url.to_dict().items())
+    urls = list(urls)
 
-    def __init__(self, layer: str, outfields: Union[str, List[str]] = "*", crs: str = DEF_CRS):
-        self.valid_layers = {
-            "wbdline": "wbdline",
-            "huc2": "2-digit hu (region)",
-            "huc4": "4-digit hu (subregion)",
-            "huc6": "6-digit hu (basin)",
-            "huc8": "8-digit hu  (subbasin)",
-            "huc10": "10-digit hu (watershed)",
-            "huc12": "12-digit hu (subwatershed)",
-            "huc14": "14-digit hu",
-            "huc16": "16-digit hu",
-        }
-        _layer = self.valid_layers.get(layer)
-        if _layer is None:
-            raise InputValueError("layer", list(self.valid_layers))
-        super().__init__(ServiceURL().restful.wbd, _layer, outfields, crs)
+    root = Path(soil_dir)
+    root.mkdir(exist_ok=True, parents=True)
+    files = [root.joinpath(f) for f in _files]
+    _ = ogc.streaming_download(urls, fnames=files)
+
+    def get_tif(file: Path) -> xr.DataArray:
+        """Get the .tif file from a zip file."""
+        with zipfile.ZipFile(file) as z:
+            try:
+                fname = next(f.filename for f in z.filelist if f.filename.endswith(".tif"))
+            except StopIteration:
+                raise ZeroMatchedError
+            ds = rxr.open_rasterio(io.BytesIO(z.read(fname)))
+            ds = cast("xr.DataArray", ds)
+            if "band" in ds.dims:
+                ds = ds.squeeze("band", drop=True)
+            ds.name = valid_props[file.stem.split("_")[0]]["name"]
+            ds.attrs["units"] = valid_props[file.stem.split("_")[0]]["units"]
+            ds.attrs["long_name"] = valid_props[file.stem.split("_")[0]]["long_name"]
+            return ds
+
+    soil = xr.merge((get_tif(f) for f in files), combine_attrs="drop_conflicts")
+    _ = soil.attrs.pop("_FillValue", None)
+    _ = soil.attrs.pop("units", None)
+    _ = soil.attrs.pop("long_name", None)
+    return soil
+
+
+def soil_gnatsgo(layers: list[str] | str, geometry: GTYPE, crs: CRSTYPE = 4326) -> xr.Dataset:
+    """Get US soil data from the gNATSGO dataset.
+
+    Notes
+    -----
+    This function uses Microsoft's Planetary Computer service to get the data.
+    The dataset's description and its suppoerted soil properties can be found at:
+    https://planetarycomputer.microsoft.com/dataset/gnatsgo-rasters
+
+    Parameters
+    ----------
+    layers : list of str or str
+        Target layer(s). Available layers can be found at the dataset's website
+        `here <https://planetarycomputer.microsoft.com/dataset/gnatsgo-rasters>`__.
+    geometry : Polygon, MultiPolygon, or tuple of length 4
+        Geometry or bounding box of the region of interest.
+    crs : int, str, or pyproj.CRS, optional
+        The input geometry CRS, defaults to ``epsg:4326``.
+
+    Returns
+    -------
+    xarray.Dataset
+        Requested soil properties.
+    """
+    try:
+        import planetary_computer
+        import pystac_client
+    except ImportError as ex:
+        raise DependencyError("soil_gnatsgo", ["pystac-client", "planetary-computer"]) from ex
+
+    catalog = pystac_client.Client.open(
+        "https://planetarycomputer.microsoft.com/api/stac/v1",
+        modifier=planetary_computer.sign_inplace,
+    )
+    bounds = geoutils.geo2polygon(geometry, crs, 4326).bounds
+    search = catalog.search(collections=["gnatsgo-rasters"], bbox=bounds)
+    lyr_href = tlz.merge_with(
+        set, ({n: a.href for n, a in i.assets.items()} for i in search.items())
+    )
+    lyrs = [layers.lower()] if isinstance(layers, str) else map(str.lower, layers)
+
+    def get_layer(lyr: str) -> xr.DataArray:
+        data = xr.open_mfdataset(lyr_href[lyr], engine="rasterio").band_data
+        if "band" in data.dims:
+            data = data.squeeze("band", drop=True)
+        data.name = lyr
+        return data  # type: ignore[no-any-return]
+
+    with dask.config.set({"array.slicing.split_large_chunks": True}):
+        ds = xr.merge((get_layer(lyr) for lyr in lyrs), combine_attrs="drop_conflicts")
+        poly = geoutils.geo2polygon(geometry, crs, ds.rio.crs)
+        ds = geoutils.xarray_geomask(ds, poly, ds.rio.crs)
+        _ = ds.attrs.pop("_FillValue", None)
+        _ = ds.attrs.pop("units", None)
+        _ = ds.attrs.pop("long_name", None)
+    return ds
